import loopy as lp
import numpy as np
import pymbolic.primitives as prim
from dune.perftool.loopy.temporary import DuneTemporaryVariable
from dune.perftool.loopy.symbolic import substitute
from dune.perftool.loopy.vcl import get_vcl_type_size
from dune.perftool.options import get_option
from dune.perftool.pdelab.argument import PDELabAccumulationFunction
from dune.perftool.pdelab.geometry import world_dimension
from dune.perftool.tools import get_pymbolic_indices


def add_vcl_temporaries(knl):
    vector_alias = [a for a in knl.arg_dict if a.endswith('alias')]

    # add new temporaries for vectors
    # hope one read insn doesn't have two different reads from the same temporary
    new_vec_temporaries = dict()
    for alias in vector_alias:
        vector_name = alias.replace('alias','vec')
        new_vec_temporaries[vector_name] = DuneTemporaryVariable(vector_name, dtype=np.float64,
                                                                 shape=(get_vcl_type_size(np.float64),), managed=True,
                                                                 scope=lp.temp_var_scope.PRIVATE, dim_tags=('vec',))

    return knl.copy(temporary_variables=dict(**knl.temporary_variables, **new_vec_temporaries))


def add_vcl_accum_insns(knl, iname_inner, iname_outer):
    from loopy.match import Tagged
    accum_insns = set(lp.find_instructions(knl, Tagged('accum')))

    new_insns = []
    vng = knl.get_var_name_generator()
    idg = knl.get_instruction_id_generator()
    new_vec_temporaries = dict()
    vcl_size = get_vcl_type_size(np.float64)
    for insn in knl.instructions:
        # somehow CInstructions are not hashable....
        if isinstance(insn, lp.MultiAssignmentBase) and insn in accum_insns:
            # write accum expr as "r = expr + r"
            expr_without_r = prim.Sum(tuple(e for e in insn.expression.children if not e == insn.assignee))

            inames_micro = set((i for i in insn.within_inames if i.startswith('micro')))
            iname_ix = next((i for i in inames_micro if i.endswith("_x")))

            # need inames for head and tail handling a priori
            from loopy.match import Not, All
            replace_head_inames = dict()
            replace_tail_inames = dict()
            for iname in inames_micro - frozenset({iname_ix}):
                head = iname+'_head'
                tail = iname+'_tail'
                replace_head_inames[iname] = prim.Variable(head)
                replace_tail_inames[iname] = prim.Variable(tail)
                knl = lp.duplicate_inames(knl, iname, Not(All()), suffix='_tail')
                knl = lp.duplicate_inames(knl, iname, Not(All()), suffix='_head')
            inames_head = frozenset((var.name for var in replace_head_inames.values()))
            inames_tail = frozenset((var.name for var in replace_tail_inames.values()))

            # erstelle a[iy] und b
            identifier_left = vng('left_node')
            identifier_right = vng('right_node')
            new_vec_temporaries[identifier_left] = DuneTemporaryVariable(identifier_left, dtype=np.float64,
                                                                      shape=(2,)*(world_dimension()-1)
                                                                            +(vcl_size,),
                                                                      managed=True, scope=lp.temp_var_scope.PRIVATE,
                                                                      dim_tags=('f',)*(world_dimension()-1)+('vec',))
            new_vec_temporaries[identifier_right] = DuneTemporaryVariable(identifier_right, dtype=np.float64,
                                                                      shape=(vcl_size,), managed=True,
                                                                      scope=lp.temp_var_scope.PRIVATE, dim_tags=('vec',))

            var_left = prim.Subscript(prim.Variable(identifier_left),
                                   tuple(prim.Variable(i) for i in sorted(inames_micro - frozenset({iname_ix})))
                                   +(prim.Variable(iname_inner),))
            var_right = prim.Subscript(prim.Variable(identifier_right), (prim.Variable(iname_inner),))

            # init a
            id_init_a = idg('insn_init_'+identifier_left)
            new_insns.append(lp.Assignment(assignee=substitute(var_left,replace_head_inames),
                                           expression=0,
                                           id=id_init_a,
                                           within_inames=(insn.within_inames-frozenset({iname_outer})-inames_micro)
                                                         | inames_head,
                                           tags=frozenset({'head'})
                                           )
                             )

            # setze werte für a und b
            expr_right = substitute(expr_without_r, {iname_ix:1})
            expr_left = prim.Sum((substitute(expr_without_r, {iname_ix:0}), var_left))

            id_set_left = idg('insn_'+identifier_left)
            id_set_right = idg('insn_'+identifier_right)
            new_insns.append(lp.Assignment(assignee=var_right,
                                           expression=expr_right,
                                           id=id_set_right,
                                           depends_on=insn.depends_on,
                                           within_inames=insn.within_inames-frozenset({iname_ix}),
                                           )
                             )
            new_insns.append(lp.Assignment(assignee=var_left,
                                           expression=expr_left,
                                           id=id_set_left,
                                           depends_on=insn.depends_on|frozenset({id_init_a}),
                                           within_inames=insn.within_inames-frozenset({iname_ix}),
                                           )
                             )

            # r+=a[iy]
            id_accum = idg('insn_mod_accum')
            expr_accum = prim.Sum((var_left,
                                   prim.Call(prim.Variable('permute{}d<-1,{}>'.format(vcl_size,','.join(map(str,range(vcl_size-1))))),
                                             (var_right,)),
                                   substitute(insn.assignee, {iname_ix:0})))
            new_insns.append(lp.Assignment(assignee=substitute(insn.assignee,{iname_ix:0}),
                                           expression=expr_accum,
                                           id=id_accum,
                                           depends_on=insn.depends_on|frozenset({id_set_left,id_init_a,id_set_right}),
                                           within_inames=insn.within_inames-frozenset({iname_ix}),
                                           tags=frozenset({'accum'})
                                           )
                             )
            # a[iy] = permute
            id_permute = idg('insn_permute')
            expr_permute = prim.Call(prim.Variable('permute{}d<3,{}>'.format(vcl_size,','.join(['-1']*(vcl_size-1)))),
                                     (var_right,))
            new_insns.append(lp.Assignment(assignee=var_left,
                                           expression=expr_permute,
                                           id=id_permute,
                                           depends_on=insn.depends_on|frozenset({id_set_left,id_init_a,id_set_right,
                                                                                 id_accum}),
                                           within_inames=insn.within_inames-frozenset({iname_ix})
                                           )
                             )

            # tail handling, uses tail alias
            id_accum_tail = idg('insn_accum_tail')
            subst_map = {iname_inner: 0, iname_outer: get_option("number_of_blocks")//vcl_size,
                         iname_ix: 0, insn.assignee_name: prim.Variable(insn.assignee_name+'_tail'),
                         **replace_tail_inames}
            assignee_tail = substitute(insn.assignee, subst_map)
            expr_tail = prim.Sum((substitute(var_left, {iname_inner: 0, **replace_tail_inames}), assignee_tail))

            new_insns.append(lp.Assignment(assignee=assignee_tail,
                                           expression=expr_tail,
                                           id=id_accum_tail,
                                           depends_on=frozenset({id_accum,id_permute, id_set_left, id_init_a}),
                                           within_inames=(insn.within_inames
                                                          - frozenset({iname_inner, iname_outer}) - inames_micro)
                                                         | inames_tail,
                                           tags=frozenset({'tail'})))
        else:
            new_insns.append(insn)

    return knl.copy(instructions=new_insns,
                    temporary_variables=dict(**knl.temporary_variables, **new_vec_temporaries))


def add_vcl_access(knl, iname_inner):
    from loopy.match import Reads, Tagged
    accum_insns = set((insn.id for insn in lp.find_instructions(knl, Tagged('accum'))))
    read_insns = set((insn.id for insn in lp.find_instructions(knl, Reads('*alias'))))

    from loopy.symbolic import CombineMapper
    from loopy.symbolic import IdentityMapper
    class AliasIndexCollector(CombineMapper, IdentityMapper):
        def combine(self, values):
            return sum(values, tuple())

        def map_constant(self, expr):
            return tuple()

        map_variable = map_constant
        map_function_symbol = map_constant

        def map_subscript(self, expr):
            if expr.aggregate.name.endswith('alias'):
                return expr.aggregate, expr.index_tuple
            else:
                return tuple()

    # add load instructions if the vector is read, based on the read instruction
    # TODO brauche mehrere vectoren, falls in einer insn von einem alias mit unterschiedlichen idx gelesen wird
    # muss dafür eigentlich nur den namen des vectors anpassen
    idg = knl.get_instruction_id_generator()
    aic = AliasIndexCollector()
    load_insns = []
    read_dependencies = dict()
    for id in read_insns:
        insn = knl.id_to_insn[id]

        alias, index = aic(insn.expression)
        name_alias = alias.name
        name_vec = name_alias.replace('alias', 'vec')

        # compute index without vec iname
        strides = tuple(tag.stride for tag in knl.arg_dict[name_alias].dim_tags)
        index = prim.Sum(tuple(prim.Product((i,s)) for i,s in zip(index, strides) if i!=0 and i.name != iname_inner))

        # add load instruction
        load_id = idg('insn_'+name_vec+'_load')
        call_load = prim.Call(prim.Variable(name_vec+'.load'), (prim.Sum((prim.Variable(name_alias), index)),))
        load_insns.append(lp.CallInstruction(assignees=(),expression=call_load,
                                             id=load_id,within_inames=insn.within_inames|insn.reduction_inames(),))
        read_dependencies.setdefault(id, set())
        read_dependencies[id].add(load_id)

    # add store instructions if the vector is written, based on the write instruction
    store_insns = []
    for id in accum_insns:
        insn = knl.id_to_insn[id]

        alias, index = aic(insn.expression)
        name_alias = alias.name
        name_vec = name_alias.replace('alias', 'vec')

        # flat index without vec iname
        strides = tuple(tag.stride for tag in knl.arg_dict[name_alias].dim_tags)
        index = prim.Sum(tuple(prim.Product((i,s)) for i,s in zip(index, strides) if i!=0 and i.name != iname_inner))

        # add store instruction
        store_id = idg('insn_'+name_vec+'_store')
        call_store = prim.Call(prim.Variable(name_vec+'.store'), (prim.Sum((prim.Variable(name_alias), index)),))
        store_insns.append(lp.CallInstruction(assignees=(),expression=call_store,
                                              id=store_id,within_inames=insn.within_inames,
                                              depends_on=insn.depends_on
                                                         | frozenset({id})
                                                         | read_dependencies[id],
                                              ))

    # replace alias with vcl vector, except for accumulation assignee
    vector_alias = [a for a in knl.arg_dict if a.endswith('alias')]
    dim = world_dimension()
    dim_names = ["x","y","z"] + [str(i) for i in range(4,dim+1)]
    knl_without_cinsn = knl.copy(instructions=[insn for insn in knl.instructions
                                               if not isinstance(insn, lp.CInstruction)])
    for alias in vector_alias:
        parameters = 'ex_o,ex_i,' + ','.join(['e'+d for d in dim_names[1:dim]])\
                     + ',ix,' + ','.join(['i'+d for d in dim_names[1:dim]])
        knl_without_cinsn = lp.extract_subst(knl_without_cinsn, alias+'_subst', '{}[{}]'.format(alias, parameters),
                                             parameters=parameters)
        new_subst = knl_without_cinsn.substitutions.copy()
        rule = new_subst[alias+'_subst']
        rule.expression = prim.Subscript(prim.Variable(alias.replace('alias','vec')), (prim.Variable('ex_i'),))
        knl_without_cinsn = knl_without_cinsn.copy(substitutions=new_subst)

    from loopy.match import All
    knl_without_cinsn = lp.expand_subst(knl_without_cinsn, All())
    knl = knl_without_cinsn.copy(instructions=knl_without_cinsn.instructions
                                              + [insn for insn in knl.instructions
                                                 if isinstance(insn, lp.CInstruction)])

    # add store and load dependencies and set right accumulation assignee
    new_insns = []
    for insn in knl.instructions:
        if insn.id not in read_insns | accum_insns:
            new_insns.append(insn)
        else:
            if insn.id in accum_insns:
                assignee_alias = insn.assignee
                assignee_vec = next((expr for expr in insn.expression.children
                                     if isinstance(expr, prim.Subscript)
                                     and expr.aggregate.name.replace('vec','alias') == assignee_alias.aggregate.name))
                new_insns.append(insn.copy(assignee=assignee_vec,
                                           depends_on=insn.depends_on | read_dependencies[insn.id]))
            else:
                new_insns.append(insn.copy(depends_on=insn.depends_on | read_dependencies[insn.id]))

    return knl.copy(instructions=new_insns+load_insns+store_insns)


def find_accumulation_inames(knl):
    inames = set()
    for insn in knl.instructions:
        if any((n.startswith('r_') and n.endswith('alias') for n in insn.write_dependency_names())):
            inames |= insn.within_inames

    inames = set((i for i in inames if i.startswith('micro') and not i.endswith('_x')))

    return inames


def vectorize_micro_elements(knl):
    if "subel_x" in knl.all_inames():
        vcl_size = get_vcl_type_size(np.float64)
        knl = lp.split_iname(knl,"subel_x", vcl_size, inner_tag='vec')
        array_alias = [a for a in knl.arg_dict.keys() if a.endswith('alias') or a.endswith('tail')]
        knl = lp.split_array_axis(knl, array_alias, 0, vcl_size)

        knl = add_vcl_temporaries(knl)
        knl = add_vcl_accum_insns(knl, 'subel_x_inner', 'subel_x_outer')
        knl = add_vcl_access(knl, 'subel_x_inner')
    return knl
