""" Generators for basis evaluations """

from dune.perftool.generation import (backend,
                                      class_member,
                                      get_backend,
                                      include_file,
                                      instruction,
                                      kernel_cached,
                                      temporary_variable,
                                      )
from dune.perftool.options import (option_switch,
                                   get_option
                                   )
from dune.perftool.pdelab.spaces import (lfs_iname,
                                         lfs_inames,
                                         name_leaf_lfs,
                                         name_lfs,
                                         name_lfs_bound,
                                         type_gfs,
                                         type_leaf_gfs,
                                         )
from dune.perftool.pdelab.geometry import (component_iname,
                                           world_dimension,
                                           name_jacobian_inverse_transposed,
                                           to_cell_coordinates,
                                           )
from dune.perftool.pdelab.localoperator import (lop_template_ansatz_gfs,
                                                lop_template_test_gfs,
                                                )
from dune.perftool.tools import (get_pymbolic_basename,
                                 get_pymbolic_indices,
                                 )
from dune.perftool.pdelab.driver import FEM_name_mangling
from dune.perftool.pdelab.restriction import restricted_name
from dune.perftool.pdelab.driver import (isPk,
                                         isQk,
                                         isDG)

from pymbolic.primitives import Product, Subscript, Variable

from ufl import MixedElement

from loopy import Reduction


@backend(interface="typedef_localbasis")
@class_member(classtag="operator")
def typedef_localbasis(element, name):
    basis_type = "{}::Traits::FiniteElementMap::Traits::FiniteElementType::Traits::LocalBasisType".format(type_leaf_gfs(element))
    return "using {} = typename {};".format(name, basis_type)


def type_localbasis(element):
    if isPk(element):
        name = "P{}_LocalBasis".format(element._degree)
    elif isQk(element):
        name = "Q{}_LocalBasis".format(element._degree)
    elif isDG(element):
        name = "DG{}_LocalBasis".format(element._degree)
    else:
        raise NotImplementedError("Element type not known in code generation")
    get_backend("typedef_localbasis", selector=option_switch("blockstructured"))(element, name)
    return name


def type_localbasis_cache(element):
    return "LocalBasisCacheWithoutReferences<{}>".format(type_localbasis(element))


@class_member(classtag="operator")
def define_localbasis_cache(element, name):
    include_file("dune/perftool/localbasiscache.hh", filetag="operatorfile")
    t = type_localbasis_cache(element)
    return "{} {};".format(t, name)


def name_localbasis_cache(element):
    name = "cache_{}".format(FEM_name_mangling(element))
    define_localbasis_cache(element, name)
    return name


def declare_cache_temporary(element, restriction, which):
    t_cache = type_localbasis_cache(element)
    lfs = name_leaf_lfs(element, restriction)

    def decl(name, shape, shape_impl):
        return "typename {}::{}ReturnType {};".format(t_cache,
                                                      which,
                                                      name,
                                                      )
    return decl


@backend(interface="evaluate_basis")
@kernel_cached
def evaluate_basis(leaf_element, name, restriction):
    lfs = name_leaf_lfs(leaf_element, restriction)
    temporary_variable(name, shape=(name_lfs_bound(lfs), 1), decl_method=declare_cache_temporary(leaf_element, restriction, 'Function'))
    cache = name_localbasis_cache(leaf_element)
    qp = get_backend("qp_in_cell")(restriction)
    instruction(inames=get_backend("quad_inames")(),
                code='{} = {}.evaluateFunction({}, {}.finiteElement().localBasis());'.format(name,
                                                                                             cache,
                                                                                             str(qp),
                                                                                             lfs,
                                                                                             ),
                assignees=frozenset({name}),
                read_variables=frozenset({get_pymbolic_basename(qp)}),
                )


def pymbolic_basis(leaf_element, restriction, number, context=''):
    assert not isinstance(leaf_element, MixedElement)
    name = "phi_{}".format(FEM_name_mangling(leaf_element))
    name = restricted_name(name, restriction)
    evaluate_basis(leaf_element, name, restriction)
    iname, = lfs_inames(leaf_element, restriction, number, context=context)

    return Subscript(Variable(name), (Variable(iname), 0))


@backend(interface="evaluate_grad")
@kernel_cached
def evaluate_reference_gradient(leaf_element, name, restriction):
    lfs = name_leaf_lfs(leaf_element, restriction)
    temporary_variable(name, shape=(name_lfs_bound(lfs), 1, world_dimension()), decl_method=declare_cache_temporary(leaf_element, restriction, 'Jacobian'))
    cache = name_localbasis_cache(leaf_element)
    qp = get_backend("qp_in_cell")(restriction)
    instruction(inames=get_backend("quad_inames")(),
                code='{} = {}.evaluateJacobian({}, {}.finiteElement().localBasis());'.format(name,
                                                                                             cache,
                                                                                             str(qp),
                                                                                             lfs,
                                                                                             ),
                assignees=frozenset({name}),
                read_variables=frozenset({get_pymbolic_basename(qp)}),
                )


def pymbolic_reference_gradient(leaf_element, restriction, number, context=''):
    assert not isinstance(leaf_element, MixedElement)
    name = "js_{}".format(FEM_name_mangling(leaf_element))
    name = restricted_name(name, restriction)
    evaluate_reference_gradient(leaf_element, name, restriction)
    iname, = lfs_inames(leaf_element, restriction, number, context=context)

    return Subscript(Variable(name), (Variable(iname), 0))


def shape_as_pymbolic(shape):
    def _shape_as_pymbolic(s):
        if isinstance(s, str):
            return Variable(s)
        else:
            return s
    return tuple(_shape_as_pymbolic(s) for s in shape)


@kernel_cached
def evaluate_coefficient(visitor, element, name, container, restriction, index):
    sub_element = element
    if isinstance(element, MixedElement):
        sub_element = element.extract_component(index)[1]

    from ufl import FiniteElement
    assert isinstance(sub_element, FiniteElement)

    temporary_variable(name, shape=(), managed=True)

    lfs = name_lfs(element, restriction, index)
    basis = visitor.interface.pymbolic_basis(sub_element, restriction, 0, context='trial')
    basisindex = get_pymbolic_indices(basis)[:-1]
    if get_option("blockstructured"):
        from dune.perftool.blockstructured.argument import pymbolic_coefficient
        coeff = pymbolic_coefficient(container, lfs, sub_element, basisindex)
    else:
        from dune.perftool.pdelab.argument import pymbolic_coefficient
        coeff = pymbolic_coefficient(container, lfs, basisindex[0])

    assignee = Variable(name)

    reduction_expr = Product((coeff, basis))

    instruction(expression=Reduction("sum", basisindex, reduction_expr, allow_simultaneous=True),
                assignee=assignee,
                forced_iname_deps=frozenset(get_backend("quad_inames")()),
                forced_iname_deps_is_final=True,
                )


@kernel_cached
def evaluate_coefficient_gradient(visitor, element, name, container, restriction, index):
    sub_element = element
    if isinstance(element, MixedElement):
        sub_element = element.extract_component(index)[1]
    from ufl import FiniteElement
    assert isinstance(sub_element, FiniteElement)

    temporary_variable(name, shape=(element.cell().geometric_dimension(),), managed=True)

    dimindex = component_iname(count=0)

    lfs = name_lfs(element, restriction, index)
    basis = visitor.interface.pymbolic_reference_gradient(sub_element, restriction, 0, context='trialgrad')
    basisindex = get_pymbolic_indices(basis)[:-1]
    from dune.perftool.tools import maybe_wrap_subscript
    basis = maybe_wrap_subscript(basis, Variable(dimindex))

    if get_option("blockstructured"):
        from dune.perftool.blockstructured.argument import pymbolic_coefficient
        coeff = pymbolic_coefficient(container, lfs, sub_element, basisindex)
    else:
        from dune.perftool.pdelab.argument import pymbolic_coefficient
        coeff = pymbolic_coefficient(container, lfs, basisindex[0])

    assignee = Subscript(Variable(name), (Variable(dimindex),))

    reduction_expr = Product((coeff, basis))

    instruction(expression=Reduction("sum", basisindex, reduction_expr, allow_simultaneous=True),
                assignee=assignee,
                forced_iname_deps=frozenset(get_backend("quad_inames")()).union(frozenset({dimindex})),
                forced_iname_deps_is_final=True
                )
